/*
 * Copyright (C) 2023 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.eingang.formsolutions;

import java.io.File;
import java.util.UUID;
import java.util.function.Supplier;

import org.apache.commons.lang3.exception.ExceptionUtils;
import org.apache.logging.log4j.CloseableThreadContext;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.ws.server.endpoint.annotation.Endpoint;
import org.springframework.ws.server.endpoint.annotation.PayloadRoot;
import org.springframework.ws.server.endpoint.annotation.RequestPayload;
import org.springframework.ws.server.endpoint.annotation.ResponsePayload;

import de.ozgcloud.common.binaryfile.TempFileUtils;
import de.ozgcloud.eingang.semantik.SemantikAdapter;
import lombok.extern.log4j.Log4j2;

@Endpoint
@Log4j2
public class SendFormEndpoint {

	public static final String JSON_FIELD = "json";

	@Autowired
	private FormSolutionsRequestMapper requestMapper;
	@Autowired
	private SemantikAdapter semantikAdapter;

	private static final String REQUEST_ID_KEY = "requestId";

	@PayloadRoot(namespace = WebServiceConfiguration.NAMESPACE_URI, localPart = "Request")
	@ResponsePayload
	public Response receiveForm(@RequestPayload Request request) {
		return doSurroundOn(() -> handleRequest(request));
	}

	private Response handleRequest(Request request) {
		try {
			semantikAdapter.processFormData(requestMapper.map(writeRequestJsonToFile(request.getJSON())));
			return buildSuccessResponse();
		} catch (Exception e) {
			LOG.error("Error on processing FormSolutions Formdata.", e);
			return ExceptionUtils.rethrow(e);
		}
	}

	private Response doSurroundOn(Supplier<Response> requestHandler) {
		UUID requestId = UUID.randomUUID();
		try (CloseableThreadContext.Instance ctc = CloseableThreadContext.put(REQUEST_ID_KEY, requestId.toString())) {
			LOG.info("START of Request with ID '{}'.", requestId);
			return requestHandler.get();
		} finally {
			LOG.info("END of Request with ID '{}'", requestId);
		}

	}

	private File writeRequestJsonToFile(String json) {
		return TempFileUtils.writeTmpFile(json);
	}

	private Response buildSuccessResponse() {
		LOG.debug("Successful processed data");

		var response = new Response();
		response.setStatus("OK");
		return response;
	}
}