/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.eingang.intelliform;

import static de.ozgcloud.eingang.intelliform.AttachmentTestFactory.*;
import static de.ozgcloud.eingang.intelliform.DepositDataTestFactory.*;
import static java.util.Collections.*;
import static org.assertj.core.api.Assertions.*;
import static org.mockito.Mockito.*;

import java.io.IOException;
import java.io.StringReader;
import java.nio.charset.Charset;
import java.util.List;

import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;

import org.apache.commons.io.FileUtils;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.ArgumentCaptor;
import org.mockito.Captor;
import org.mockito.Spy;
import org.w3c.dom.Document;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;

import de.ozgcloud.eingang.common.errorhandling.TechnicalException;
import de.ozgcloud.eingang.common.formdata.FormData;
import de.ozgcloud.eingang.common.formdata.IncomingFile;
import de.ozgcloud.eingang.common.formdata.IncomingFileGroup;
import lombok.SneakyThrows;

class DepositDataMapperTest {

	@Spy
	private DepositDataMapper mapper;

	@DisplayName("map to form data")
	@Nested
	class TestMapToFormData {

		@Captor
		private ArgumentCaptor<Attachment> attachmentArgumentCaptor;

		private DepositData depositData;

		@DisplayName("with normal attachments")
		@Nested
		class TestWithNormalAttachments {
			@BeforeEach
			void mock() {
				depositData = DepositDataTestFactory.create(ATTACHMENTS);
			}

			@DisplayName("should throw technical exception if primary xml link is incorrect")
			@Test
			void shouldThrowTechnicalExceptionIfPrimaryXmlLinkIsIncorrect() {
				depositData.setPrimaryDataAttachmentId("incorrect");

				assertThatThrownBy(TestMapToFormData.this::doMapping)
						.isInstanceOf(TechnicalException.class);
			}

			@DisplayName("should use map to incoming file method")
			@Test
			void shouldUseMapToIncomingFileMethod() {
				doMapping();
				verify(mapper, times(ATTACHMENTS.size())).mapAttachmentToIncomingFile(attachmentArgumentCaptor.capture());

				assertThat(attachmentArgumentCaptor.getAllValues()).isEqualTo(ATTACHMENTS);
			}

			@DisplayName("should return with representations")
			@Test
			void shouldReturnWithRepresentations() {
				var formData = doMapping();

				var incomingFileIds = formData.getRepresentations().stream().map(IncomingFile::getVendorId).toList();
				assertThat(incomingFileIds).containsExactly(XML_ATTACHMENT_ID, PDF_ATTACHMENT_ID);
			}

			@DisplayName("should return with one attachment")
			@Test
			void shouldReturnWithOneAttachment() {
				var formData = doMapping();

				var incomingFileIds = formData.getAttachments().stream()
						.flatMap(group -> group.getFiles().stream())
						.map(IncomingFile::getVendorId)
						.toList();
				assertThat(incomingFileIds).containsExactly(PNG_ATTACHMENT_ID);
			}

			@DisplayName("should return with attachment group name")
			@Test
			void shouldReturnWithAttachmentGroupName() {
				var formData = doMapping();

				var incomingFileIds = formData.getAttachments().stream()
						.map(IncomingFileGroup::getName)
						.toList();
				assertThat(incomingFileIds).containsExactly("Upload1");
			}

			@DisplayName("should return with number of representations")
			@Test
			void shouldReturnWithNumberOfRepresentations() {
				var formData = doMapping();

				assertThat(formData.getNumberOfRepresentations()).isEqualTo(2);
			}
		}

		@DisplayName("with duplicate keys")
		@Nested
		class TestWithDuplicateKeys {
			@BeforeEach
			void mock() {
				depositData = DepositDataTestFactory.create(List.of(
						withEmptyName(createXmlDaten()),
						createXmlDaten(),
						withEmptyName(createPdf()),
						createPdf(),
						withEmptyName(createPng()),
						createPng()
				));
			}

			private Attachment withEmptyName(Attachment attachment) {
				attachment.setName("");
				return attachment;
			}

			@DisplayName("should keep last entry for representations")
			@Test
			void shouldKeepLastEntryForRepresentations() {
				var formData = doMapping();

				var representationFiles = formData.getRepresentations();
				assertThat(getAttachmentVendorIds(representationFiles)).containsExactly(XML_ATTACHMENT_ID, PDF_ATTACHMENT_ID);
				assertThat(getAttachmentFileNames(representationFiles)).containsExactly(XML_NAME, PDF_ATTACHMENT_NAME);
			}

			@DisplayName("should keep last entry for attachments")
			@Test
			void shouldKeepLastEntryForAttachments() {
				var formData = doMapping();

				var attachmentFiles = formData.getAttachments().stream()
						.map(IncomingFileGroup::getFiles)
						.flatMap(List::stream)
						.toList();
				assertThat(getAttachmentVendorIds(attachmentFiles)).containsExactly(PNG_ATTACHMENT_ID);
				assertThat(getAttachmentFileNames(attachmentFiles)).containsExactly(PNG_ATTACHMENT_NAME);
			}

			private List<String> getAttachmentFileNames(List<IncomingFile> incomingFileList) {
				return incomingFileList.stream()
						.map(IncomingFile::getName)
						.toList();
			}

			private List<String> getAttachmentVendorIds(List<IncomingFile> incomingFileList) {
				return incomingFileList.stream()
						.map(IncomingFile::getVendorId)
						.toList();
			}
		}

		@DisplayName("with many attachments")
		@Nested
		class TestWithManyAttachments {
			@BeforeEach
			void mock() {
				depositData = DepositDataTestFactory.create(MANY_ATTACHMENTS);
			}

			@DisplayName("should return with representations")
			@Test
			void shouldReturnWithRepresentations() {
				var formData = doMapping();

				var incomingFileIds = formData.getRepresentations().stream()
						.map(IncomingFile::getVendorId)
						.toList();
				assertThat(incomingFileIds).containsExactly(
						XML_ATTACHMENT_ID,
						XML_ROHFORM_ATTACHMENT_ID,
						XML_ORIGINALFORM_ATTACHMENT_ID
				);
			}

			@DisplayName("should return with attachment groups")
			@Test
			void shouldReturnWithAttachmentGroups() {
				var formData = doMapping();

				var incomingFileIds = formData.getAttachments().stream()
						.flatMap(group -> group.getFiles().stream())
						.map(IncomingFile::getVendorId)
						.toList();
				assertThat(incomingFileIds).containsExactlyInAnyOrder(
						DOCX1_ATTACHMENT_ID,
						PDF_ATTACHMENT_ID,
						DOCX2_ATTACHMENT_ID,
						DOCX3_ATTACHMENT_ID,
						DOCX4_ATTACHMENT_ID,
						DOCX5_ATTACHMENT_ID,
						DOCX6_ATTACHMENT_ID,
						DOCX7_ATTACHMENT_ID,
						DOCX8_ATTACHMENT_ID,
						PDF2_ATTACHMENT_ID,
						ODT_ATTACHMENT_ID,
						JPG_ATTACHMENT_ID,
						PNG_ATTACHMENT_ID
				);
			}
		}

		@DisplayName("with empty attachments")
		@Nested
		class TestWithEmptyAttachments {
			@DisplayName("should throw technical exception")
			@Test
			void shouldThrowTechnicalException() {
				depositData = DepositDataTestFactory.create(emptyList());

				assertThatThrownBy(TestMapToFormData.this::doMapping)
						.isInstanceOf(TechnicalException.class);
			}
		}

		private FormData doMapping() {
			return mapper.mapToFormData(depositData);
		}

	}

	@DisplayName("map deposit attachments to sorted incoming files")
	@Nested
	class TestMapDepositAttachmentsToSortedIncomingFiles {

		@DisplayName("should fail without primaryDataAttachmentId attachment")
		@Test
		void shouldFailWithoutPrimaryDataAttachmentIdAttachment() {
			var depositData = DepositDataTestFactory.create(emptyList());

			assertThatThrownBy(() -> mapper.mapDepositAttachmentsToSortedIncomingFiles(depositData))
					.isInstanceOf(TechnicalException.class);
		}

		@DisplayName("should keep entry order")
		@Test
		void shouldKeepEntryOrder() {
			var depositData = DepositDataTestFactory.create(ATTACHMENTS);

			var incomingFileMap = mapper.mapDepositAttachmentsToSortedIncomingFiles(depositData);

			var keys = incomingFileMap.keySet().stream().toList();
			assertThat(keys).containsExactly(XML_ATTACHMENT_ID, PDF_ATTACHMENT_ID, PNG_ATTACHMENT_ID);
		}

		@DisplayName("should move primary attachment id to first position")
		@Test
		void shouldMovePrimaryAttachmentIdToFirstPosition() {
			var rohformId = "XML-daten-rohform";
			var depositData = DepositDataTestFactory.create(List.of(
					AttachmentTestFactory.createAttachment(MetaAttachment.builder()
							.id(rohformId)
							.name("XML-Daten (Rohform).xml")
							.contentType(XML_CONTENT_TYPE)
							.content("abc")
							.build()
					),
					AttachmentTestFactory.createPdf(),
					AttachmentTestFactory.createXmlDaten(),
					AttachmentTestFactory.createPng()
			));

			var incomingFileMap = mapper.mapDepositAttachmentsToSortedIncomingFiles(depositData);

			var keys = incomingFileMap.keySet().stream().toList();
			assertThat(keys).containsExactly(XML_ATTACHMENT_ID, rohformId, PDF_ATTACHMENT_ID, PNG_ATTACHMENT_ID);
		}

		@DisplayName("should keep last entry for duplicate key")
		@Test
		void shouldKeepLastEntryForDuplicateKey() {
			var depositData = DepositDataTestFactory.create(List.of(
					AttachmentTestFactory.createXmlDaten(),
					AttachmentTestFactory.createPdf(),
					AttachmentTestFactory.createXmlDaten()
			));

			var incomingFileMap = mapper.mapDepositAttachmentsToSortedIncomingFiles(depositData);

			var keys = incomingFileMap.keySet().stream().toList();
			assertThat(keys).containsExactly(XML_ATTACHMENT_ID, PDF_ATTACHMENT_ID);
		}
	}

	@DisplayName("map attachment to incoming file")
	@Nested
	class TestMapAttachmentToIncomingFile {
		private Attachment attachment;

		@BeforeEach
		void mock() {
			attachment = AttachmentTestFactory.createXmlDaten();
		}

		@DisplayName("should have ID")
		@Test
		void shouldHaveId() {
			var incomingFile = doMapping();

			assertThat(incomingFile.getId()).isNotNull();
		}

		@DisplayName("should have vendor ID")
		@Test
		void shouldHaveVendorId() {
			var incomingFile = doMapping();

			assertThat(incomingFile.getVendorId()).isEqualTo(XML_ATTACHMENT_ID);
		}

		@DisplayName("should have name")
		@Test
		void shouldHaveName() {
			var incomingFile = doMapping();

			assertThat(incomingFile.getName()).isEqualTo(XML_NAME);
		}

		@DisplayName("should have content type")
		@Test
		void shouldHaveContentType() {
			var incomingFile = doMapping();

			assertThat(incomingFile.getContentType()).isEqualTo(XML_CONTENT_TYPE);
		}

		@DisplayName("should have size")
		@Test
		void shouldHaveSize() {
			var incomingFile = doMapping();

			assertThat(incomingFile.getSize()).isEqualTo(XML_CONTENT.getBytes().length);
		}

		@DisplayName("should have file with content")
		@SneakyThrows
		@Test
		void shouldHaveFileWithContent() {
			var incomingFile = doMapping();

			var content = FileUtils.readFileToString(incomingFile.getFile(), Charset.defaultCharset());
			assertThat(content).isEqualTo(XML_CONTENT_STRING);
		}

		private IncomingFile doMapping() {
			return mapper.mapAttachmentToIncomingFile(attachment);
		}
	}

	@DisplayName("find attachment groups")
	@Nested
	class TestFindAttachmentGroups {

		private Document document;

		@BeforeEach
		void mock() {
			document = buildXMLDocument("""
					<myForm t:client-id="land">
					  <Upload1>
					    <file content-type="image/png" description="" id="VendorId3333" length="155251">Image.png</file>
					  </Upload1>
					  <Upload2>
					    <file id="VendorId1111">name1.txt</file>
					    <file id="VendorId2222">name2.txt</file>
					  </Upload2>
					  <file id="VendorIdxxxx">namex.txt</file>
					</myForm>""");
		}

		@DisplayName("should have groups with parent element name")
		@Test
		void shouldHaveGroupsWithParentElementName() {
			var attachmentGroups = mapper.findAttachmentGroups(document);

			assertThat(attachmentGroups).containsOnlyKeys("Upload1", "Upload2", "myForm");
		}

		@DisplayName("should have Upload1 group with fileId")
		@Test
		void shouldHaveUpload1GroupWithFileId() {
			var attachmentGroups = mapper.findAttachmentGroups(document);

			assertThat(attachmentGroups.get("Upload1")).containsExactly("VendorId3333");
		}

		@DisplayName("should have two fileIds in Upload2 group")
		@Test
		void shouldHaveTwoFileIdsInUpload2Group() {
			var attachmentGroups = mapper.findAttachmentGroups(document);

			assertThat(attachmentGroups.get("Upload2")).containsExactly("VendorId1111", "VendorId2222");
		}
	}

	private Document buildXMLDocument(String xmlString) {
		try {
			return DocumentBuilderFactory.newInstance()
					.newDocumentBuilder()
					.parse(new InputSource(new StringReader(xmlString)));
		} catch (ParserConfigurationException | IOException | SAXException e) {
			throw new RuntimeException(e);
		}
	}

}
