/*
 * Copyright (C) 2023 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.eingang.semantik.enginebased.afm;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.util.HashMap;
import java.util.UUID;

import org.assertj.core.data.MapEntry;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.InjectMocks;
import org.mockito.Mock;

import de.ozgcloud.eingang.common.formdata.Antragsteller;
import de.ozgcloud.eingang.common.formdata.FormData;

class AfmAntragstellerMapperTest {

	@InjectMocks
	private AfmAntragstellerMapper mapper;

	@Mock
	private AfmAntragstellerHeaderMapper antragstellerHeaderMapper;

	private FormData formData = FormData.builder().formData(AfmAntragstellerTestFactory.createFormDataMap()).build();

	@DisplayName("Parse form data")
	@Nested
	class TestParseFormData {

		@Test
		void shouldCallHeaderMapper() {
			FormData expectedFormData = mock(FormData.class);
			when(antragstellerHeaderMapper.isResponsible(any())).thenReturn(true);
			when(antragstellerHeaderMapper.parseAntragstellerData(any())).thenReturn(expectedFormData);

			var processedFormData = parseFormData(formData);

			assertThat(processedFormData).isEqualTo(expectedFormData);
		}

		@Test
		void shouldDoNothingOnNullAntragstaller() {
			var formData = FormData.builder().formData(new HashMap<>()).build();

			var parsedFormData = parseFormData(formData);

			assertThat(parsedFormData).usingRecursiveComparison().ignoringFields(AfmAntragstellerMapper.ANTRAGSTELLER).isEqualTo(formData);
		}

		@Test
		void shouldMapAntragsteller() {
			var parsedFormData = parseFormData(formData);

			assertThat(parsedFormData.getAntragsteller()).usingRecursiveComparison().ignoringFields("data")
					.isEqualTo(AfmAntragstellerTestFactory.create());
		}

		@Test
		@DisplayName("should map antragsteller when key starts with an uppercase letter")
		void shouldMapAntragstellerUppercase() {
			formData = FormData.builder().formData(AfmAntragstellerTestFactory.createFormDataMap(AfmAntragstellerMapper.ANTRAGSTELLER_UPPERCASE))
					.build();

			var parsedFormData = parseFormData(formData);

			assertThat(parsedFormData.getAntragsteller()).usingRecursiveComparison().ignoringFields("data")
					.isEqualTo(AfmAntragstellerTestFactory.create());
		}

		@Test
		@DisplayName("should map only antragsteller key when both present")
		void shouldMapOnlyOneKey() {
			var expectedAntragsteller = AfmAntragstellerTestFactory.createBuilder().anrede("anrede").vorname("firstName").nachname("lastName")
					.build();

			var parsedFormData = parseFormData(createFormData(expectedAntragsteller));

			assertThat(parsedFormData.getAntragsteller()).usingRecursiveComparison().ignoringFields("data")
					.isEqualTo(expectedAntragsteller);
		}

		private FormData createFormData(Antragsteller antragsteller) {
			var formDataMap = AfmAntragstellerTestFactory.createMutableFormDataMap(AfmAntragstellerMapper.ANTRAGSTELLER_UPPERCASE);
			var antragstelleMap = AfmAntragstellerTestFactory.createPrivateAntragstelleMap(
					MapEntry.entry(AfmAntragstellerMapper.ANREDE, antragsteller.getAnrede()),
					MapEntry.entry(AfmAntragstellerMapper.VORNAME, antragsteller.getVorname()),
					MapEntry.entry(AfmAntragstellerMapper.NACHNAME, antragsteller.getNachname())
			);
			formDataMap.put(AfmAntragstellerMapper.ANTRAGSTELLER, antragstelleMap);
			return FormData.builder().formData(formDataMap).build();
		}

		@DisplayName("map private antragsteller data")
		@Nested
		class TestMapPrivateAntragstellerData {

			@Test
			void shouldMapPostfachId() {
				var parsedFormData = parseFormData(formData);

				assertThat(parsedFormData.getAntragsteller().getPostfachId()).isEqualTo(AfmAntragstellerTestFactory.POSTFACH_ID);
			}

			@DisplayName("with mapped and not mapped value")
			@Nested
			class TestWithMappedAndNotMappedValue {

				private static final String NOT_MAPPED_FIELD = "not_mapped_value";
				private static final String NOT_MAPPED_VALUE = UUID.randomUUID().toString();

				private FormData formData;

				@BeforeEach
				void buildFormData() {
					var antragstellerMap = AfmAntragstellerTestFactory.createPrivateAntragstelleMap(
							MapEntry.entry(NOT_MAPPED_FIELD, NOT_MAPPED_VALUE));

					var formDataMap = new HashMap<String, Object>();
					formDataMap.put(AfmAntragstellerMapper.ANTRAGSTELLER, antragstellerMap);

					formData = FormData.builder().formData(formDataMap).build();
				}

				@Test
				void shouldNotContainDuplicateValues() {
					var parsedFormData = parseFormData(formData);

					assertThat(parsedFormData.getAntragsteller().getData()).doesNotContainKeys(
							AfmAntragstellerMapper.ANREDE,
							AfmAntragstellerMapper.EMAIL,
							AfmAntragstellerMapper.GEBURTSDATUM,
							AfmAntragstellerMapper.GEBURTSNAME,
							AfmAntragstellerMapper.GEBURTSORT,
							AfmAntragstellerMapper.NACHNAME,
							AfmAntragstellerMapper.VORNAME,
							AfmAntragstellerMapper.TELEFON,
							AfmAntragstellerMapper.STRASSE,
							AfmAntragstellerMapper.HAUSNUMMER,
							AfmAntragstellerMapper.ORT,
							AfmAntragstellerMapper.PLZ);
				}

				@Test
				void shouldMoveNotMappedFieldsToDataMap() {
					var parsedFormData = parseFormData(formData);

					assertThat(parsedFormData.getAntragsteller().getData()).containsEntry(NOT_MAPPED_FIELD, NOT_MAPPED_VALUE);
				}
			}
		}

		@DisplayName("remove private fields")
		@Nested
		class TestRemovePrivateFields {

			@Test
			void shouldRemoveAntragsteller() {
				var parsedFormData = parseFormData(formData);

				assertThat(parsedFormData.getFormData().get(AfmAntragstellerMapper.ANTRAGSTELLER)).isNull();
			}

			@Test
			@DisplayName("should remove Antragsteller when key starts with an uppercase latter")
			void shouldRemoveAntragstellerUppercase() {
				formData = FormData.builder().formData(AfmAntragstellerTestFactory.createFormDataMap(AfmAntragstellerMapper.ANTRAGSTELLER_UPPERCASE))
						.build();

				var parsedFormData = parseFormData(formData);

				assertThat(parsedFormData.getFormData().get(AfmAntragstellerMapper.ANTRAGSTELLER_UPPERCASE)).isNull();
			}

			@Test
			@DisplayName("should remove only 'antragsteller' key when both present")
			void shouldRemoveOnlyOneKey() {
				var parsedFormData = parseFormData(createFormData());

				assertThat(parsedFormData.getFormData().get(AfmAntragstellerMapper.ANTRAGSTELLER_UPPERCASE)).isNotNull();
			}

			private FormData createFormData() {
				var formDataMap = AfmAntragstellerTestFactory.createMutableFormDataMap(AfmAntragstellerMapper.ANTRAGSTELLER_UPPERCASE);
				formDataMap.put(AfmAntragstellerMapper.ANTRAGSTELLER, AfmAntragstellerTestFactory.createPrivateAntragstelleMap());
				return FormData.builder().formData(formDataMap).build();
			}
		}
	}

	@Nested
	class TestMapFirmAntragstellerData {

		@Test
		void shouldMapFirmAntragstellerData() {
			var parsedFormData = parseFormData(createFormData());

			assertThat(parsedFormData.getAntragsteller()).usingRecursiveComparison()
					.ignoringFields("data")
					.isEqualTo(AfmAntragstellerTestFactory.createFirm());
		}

		private FormData createFormData() {
			var formDataMap = AfmAntragstellerTestFactory.createFormDataMap();
			formDataMap.put(AfmAntragstellerMapper.ANTRAGSTELLER, AfmAntragstellerTestFactory.createFirmAntragstelleMap());
			return FormData.builder().formData(formDataMap).build();
		}
	}

	private FormData parseFormData(FormData formData) {
		return mapper.parseFormData(formData);
	}
}
