/*
 * Copyright (C) 2023 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.eingang.semantik.enginebased.afm;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.Mockito.*;

import java.io.File;
import java.util.Map;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;

import com.fasterxml.jackson.dataformat.xml.XmlMapper;

import de.ozgcloud.eingang.common.formdata.FormData;
import de.ozgcloud.eingang.common.formdata.FormDataTestFactory;
import de.ozgcloud.eingang.common.formdata.IncomingFile;
import de.ozgcloud.eingang.common.formdata.IncomingFileTestFactory;
import de.ozgcloud.eingang.common.formdata.ZustaendigeStelle;
import de.ozgcloud.eingang.common.formdata.ZustaendigeStelleTestFactory;
import lombok.SneakyThrows;

class ZustaendigeStelleMetadataMapperTest {

	@Spy
	@InjectMocks
	private ZustaendigeStelleMetadataMapper mapper;

	@Mock
	private XmlMapper xmlMapper;

	@Nested
	class TestParseBehoerdeMetadata {

		private IncomingFile behoerdeMetadataFile;

		private FormData formData;

		@BeforeEach
		void setUp() {
			behoerdeMetadataFile = IncomingFileTestFactory.createBuilder().name(ZustaendigeStelleMetadataMapper.BEHOERDE_METADATA_FILE_NAME)
					.build();
			formData = FormDataTestFactory.createBuilder()
					.clearZustaendigeStelles()
					.representation(behoerdeMetadataFile).build();
		}

		@SneakyThrows
		@Test
		void shouldCallReadBehoerdeMetadata() {
			doReturn(ZustaendigeStelleDataTestFactory.BEHOERDE_METADATA).when(mapper).readZustaendigeStelleMetadata(behoerdeMetadataFile);

			mapper.parseZustaendigeStelleData(formData);

			verify(mapper).readZustaendigeStelleMetadata(behoerdeMetadataFile);
		}

		@Test
		void shouldCallMapZustaendigeStelle() {
			var behoerdeMetadata = Map.of("key", "value");
			doReturn(behoerdeMetadata).when(mapper).readZustaendigeStelleMetadata(behoerdeMetadataFile);

			mapper.parseZustaendigeStelleData(formData);

			verify(mapper).mapZustaendigeStelle(behoerdeMetadata);
		}

		@Test
		void shouldSetParsedZustaendigeStelle() {
			doReturn(Map.of("key", "value")).when(mapper).readZustaendigeStelleMetadata(any());
			var zustaendigeStelle = ZustaendigeStelleTestFactory.create();
			doReturn(zustaendigeStelle).when(mapper).mapZustaendigeStelle(any());

			var result = mapper.parseZustaendigeStelleData(formData);

			assertThat(result.getZustaendigeStelles()).containsExactly(zustaendigeStelle);
		}
	}

	@Nested
	class TestReadBehoerdeMetadata {

		private File brokenFile;

	}

	@Nested
	class TestMapZuestaendigeStelle {

		@Test
		void shouldSetBezeichnung() {
			var zustaendigeStelle = mapZustaendigeStelle();

			assertThat(zustaendigeStelle.getBezeichnung()).isEqualTo(ZustaendigeStelleDataTestFactory.BEHOERDE_ANZEIGE_NAME);
		}

		@Test
		void shouldSetOrganisationEinheitId() {
			var zustaendigeStelle = mapZustaendigeStelle();

			assertThat(zustaendigeStelle.getOrganisationseinheitenId()).isEqualTo(ZustaendigeStelleDataTestFactory.BEHOERDE_CALLER_ID);
		}

		@Test
		void shouldSetGemeindeSchluessel() {
			var zustaendigeStelle = mapZustaendigeStelle();

			assertThat(zustaendigeStelle.getGemeindeSchluessel()).isEqualTo(ZustaendigeStelleDataTestFactory.GEMEINDE_SCHLUESSEL_BP);
		}

		@Test
		void shouldSetEmail() {
			var zustaendigeStelle = mapZustaendigeStelle();

			assertThat(zustaendigeStelle.getEmail()).isEqualTo(ZustaendigeStelleDataTestFactory.BEHOERDE_EMAIL);
		}

		@Test
		void shouldSetAmtlicherRegionalSchluessel() {
			var zustaendigeStelle = mapZustaendigeStelle();

			assertThat(zustaendigeStelle.getAmtlicherRegionalSchluessel()).isEqualTo(ZustaendigeStelleDataTestFactory.AMTLICHER_REGIONALSCHLUESSEL);
		}

		@Test
		void shouldSetHausanschriftStrasse() {
			var zustaendigeStelle = mapZustaendigeStelle();

			assertThat(zustaendigeStelle.getHausanschriftStrasse()).isEqualTo(ZustaendigeStelleDataTestFactory.BEHOERDE_HAUSANSCHRIFT_STRASSE);
		}

		@Test
		void shouldSetHausanschriftOrt() {
			var zusatendigeStelle = mapZustaendigeStelle();

			assertThat(zusatendigeStelle.getHausanschriftOrt()).isEqualTo(ZustaendigeStelleDataTestFactory.BEHOERDE_HAUSANSCHRIFT_ORT);
		}

		@Test
		void shouldSetHausanschriftPlz() {
			var zustaendigeStelle = mapZustaendigeStelle();

			assertThat(zustaendigeStelle.getHausanschriftPlz()).isEqualTo(ZustaendigeStelleDataTestFactory.BEHOERDE_HAUSANSCHRIFT_PLZ);
		}

		@Test
		void shouldSetTelefon() {
			var zusatendigeStelle = mapZustaendigeStelle();

			assertThat(zusatendigeStelle.getTelefon()).isEqualTo(ZustaendigeStelleDataTestFactory.BEHOERDE_TELEFON);
		}

		private ZustaendigeStelle mapZustaendigeStelle() {
			return mapper.mapZustaendigeStelle(ZustaendigeStelleDataTestFactory.BEHOERDE_METADATA);
		}
	}

	@Test
	void shouldApproveResponsibility() {
		var formData = FormDataTestFactory.createBuilder()
				.representation(IncomingFileTestFactory.createBuilder().name(ZustaendigeStelleMetadataMapper.BEHOERDE_METADATA_FILE_NAME)
						.build()).build();

		var isResponsible = mapper.isResponsible(formData);

		assertThat(isResponsible).isTrue();
	}
}