/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.eingang.semantik.enginebased.afm.intelliform;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.util.Collections;
import java.util.List;
import java.util.Map;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.w3c.dom.Document;
import org.w3c.dom.Node;

import de.ozgcloud.common.errorhandling.TechnicalException;

class XmlToJavaMapsMapperTest {

	@InjectMocks
	private XmlToJavaMapsMapper mapper = new XmlToJavaMapsMapper();
	@Mock
	private JsonService jsonService;

	private final Document document = mapper.parseAsW3cDocument(
			XmlToJavaMapsMapperTest.class.getResourceAsStream("/intelliform/SimpleFormDataMapperTestFile.xml"));

	@Test
	void testSimpleNode() {

		Map<String, Object> formData = mapXmlToJavaMaps();

		assertThat(formData).containsKey("simplenode");
		assertThat(formData.get("simplenode")).isInstanceOf(String.class).isEqualTo("simplenodevalue");
	}

	@Test
	void testEmptyNodeNotMapped() {

		Map<String, Object> formData = mapXmlToJavaMaps();

		assertThat(formData).doesNotContainKey("emptynode");
	}

	@Nested
	class TestNestedMapNode {

		private static final String MAIN_NODE_NAME = "nestedmapnode";

		@Test
		void nestedMapNodeShouldBeMap() {

			Map<String, Object> formData = mapXmlToJavaMaps();

			assertThat(formData).containsKey(MAIN_NODE_NAME);
			assertThat(formData.get(MAIN_NODE_NAME)).isInstanceOf(Map.class);
		}

		@Test
		@SuppressWarnings("unchecked")
		void nestedMapNodeShouldContainValues() {

			Map<String, Object> nestedMapNode = (Map<String, Object>) mapXmlToJavaMaps().get(MAIN_NODE_NAME);

			assertThat(nestedMapNode).hasSize(2)
					.containsEntry("nestedmapnode1", "nestedmapnodevalue1")
					.containsEntry("nestedmapnode2", "nestedmapnodevalue2");
		}
	}

	@Nested
	@SuppressWarnings("unchecked")
	class TestNestedListNodeWithStrings {

		private static final String MAIN_NODE_NAME = "nestedlistnodewithstrings";
		private static final String NESTED_NODE_NAME = "nestednode";

		@Test
		void shouldContainNestedListWithStringsNode() {

			Map<String, Object> formData = mapXmlToJavaMaps();

			assertThat(formData).containsKey(MAIN_NODE_NAME);
		}

		@Test
		void valuesShouldBeJavaList() {

			Map<String, Object> formData = mapXmlToJavaMaps();

			Map<String, Object> mainNode = (Map<String, Object>) formData.get(MAIN_NODE_NAME);

			assertThat(mainNode).containsKey(NESTED_NODE_NAME).isInstanceOf(Map.class);
			assertThat(mainNode.get(NESTED_NODE_NAME)).isInstanceOf(List.class);
		}

		@Test
		void shouldContainValues() {

			Map<String, Object> formData = mapXmlToJavaMaps();
			Map<String, Object> mainNode = (Map<String, Object>) formData.get(MAIN_NODE_NAME);
			List<String> nestedList = (List<String>) mainNode.get(NESTED_NODE_NAME);

			assertThat(nestedList).hasSize(3).contains("nestedlistnodevalue1", "nestedlistnodevalue2", "nestedlistnodevalue3");
		}
	}

	@Nested
	@SuppressWarnings("unchecked")
	class TestNestedListNodeWithObjects {

		private static final String MAIN_NODE_NAME = "nestedlistnodewithmap";
		private static final String NESTED_NODE_NAME = "nestednode";

		@Test
		void shouldContainNestedListWithObjectsNode() {

			Map<String, Object> formData = mapXmlToJavaMaps();

			assertThat(formData).containsKey(MAIN_NODE_NAME);
		}

		@Test
		void valuesShouldBeJavaList() {

			Map<String, Object> formData = mapXmlToJavaMaps();

			Map<String, Object> mainNode = (Map<String, Object>) formData.get(MAIN_NODE_NAME);

			assertThat(mainNode).containsKey(NESTED_NODE_NAME).isInstanceOf(Map.class);
			assertThat(mainNode.get(NESTED_NODE_NAME)).isInstanceOf(List.class);
		}

		@Test
		void shouldContainTwoElements() {

			Map<String, Object> formData = mapXmlToJavaMaps();
			List<Object> nestedList = getNestedList(formData);

			assertThat(nestedList).hasSize(2);
		}

		@Test
		void validateLevel2ValuesOfMap1() {

			Map<String, Object> formData = mapXmlToJavaMaps();
			List<Object> nestedList = getNestedList(formData);

			Map<String, Object> map1 = (Map<String, Object>) nestedList.get(0);

			assertThat(map1).hasSize(1)
					.containsEntry("level2", "nestedlistnodevalue1");
		}

		@Test
		void validateLevel2ValuesOfMap2() {

			Map<String, Object> formData = mapXmlToJavaMaps();
			List<Object> nestedList = getNestedList(formData);

			Map<String, Object> map2 = (Map<String, Object>) nestedList.get(1);

			assertThat(map2).hasSize(1);
			assertThat(map2.get("level2")).isInstanceOf(List.class);

			List<String> map21 = (List<String>) map2.get("level2");
			assertThat(map21).contains("nestedlistnodevalue1", "nestedlistnodevalue2", "nestedlistnodevalue3");
		}

		private List<Object> getNestedList(Map<String, Object> formData) {
			Map<String, Object> mainNode = (Map<String, Object>) formData.get(MAIN_NODE_NAME);
			List<Object> nestedList = (List<Object>) mainNode.get(NESTED_NODE_NAME);
			return nestedList;
		}
	}

	@Nested
	class TestParseFile {

		private static final String FILE = "file";

		@Nested
		class TestParseFileList {

			private static final String LIST_FILE_PARENT = "list_file_parent";
			private static final String LIST_FILE_PARENT_ITEM = "list_file_parent-item";

			@Test
			void shouldMapAttributesOnListFirstElement() {
				var javaMap = mapXmlToJavaMaps();

				assertThat(getListItemFile(javaMap, 0))
						.containsEntry("id", "assistants.21B483DAA2DC7900C1D1135E566D0F672CB42832")
						.containsEntry("content-type", "image/jpeg")
						.containsEntry("description", "file description for first list file")
						.containsEntry("length", "1184319")
						.containsEntry("text", "20210326_133516[1].jpg");
			}

			@Test
			void shouldMapAttributesOnListSecondElement() {
				var javaMap = mapXmlToJavaMaps();

				assertThat(getListItemFile(javaMap, 1))
						.containsEntry("id", "assistants.2F1FF737CF4F23191C1952091CB342254EE22A37")
						.containsEntry("content-type", "image/jpeg")
						.containsEntry("description", "file description for second list file")
						.containsEntry("length", "1384037")
						.containsEntry("text", "20210326_133526[1].jpg");
			}

			@SuppressWarnings("unchecked")
			private Map<String, Object> getListItemFile(Map<String, Object> map, int index) {
				var listFileParent = (Map<String, Object>) map.get(LIST_FILE_PARENT);
				var listFileParentItem = (List<?>) listFileParent.get(LIST_FILE_PARENT_ITEM);
				var firstFileItem = (Map<String, Object>) listFileParentItem.get(index);
				return (Map<String, Object>) firstFileItem.get(FILE);
			}
		}

		@Nested
		class TestParseSingleFile {

			private static final String SINGLE_FILE_PARENT = "single_file_parent";
			private static final String SINGLE_FILE_PARENT_ITEM = "single_file_parent-item";

			@Test
			void shouldMapAttributesOnSingleElement() {
				var javaMap = mapXmlToJavaMaps();

				assertThat(getSingleItemFile(javaMap))
						.containsEntry("id", "assistants.B20451156C677116B91A4EF2D1E859837B6CE510")
						.containsEntry("content-type", "image/jpeg")
						.containsEntry("description", "file description for single file")
						.containsEntry("length", "3066416")
						.containsEntry("text", "20210324_103054[1].jpg");
			}

			@SuppressWarnings("unchecked")
			private Map<String, Object> getSingleItemFile(Map<String, Object> map) {
				var singleFileParent = (Map<String, Object>) map.get(SINGLE_FILE_PARENT);
				var singleFileParentItem = (Map<String, Object>) singleFileParent.get(SINGLE_FILE_PARENT_ITEM);
				return (Map<String, Object>) singleFileParentItem.get(FILE);
			}
		}
	}

	@DisplayName("Get content value")
	@Nested
	class TestGetContentValue {

		@DisplayName("from json node")
		@Nested
		class TestJsonNode {

			private static final String VALID_EMPTY_JSON = "{}";

			@Mock
			private Node node;

			@BeforeEach
			void mockNode() {
				when(node.getNodeName()).thenReturn(XmlToJavaMapsMapper.REST_RESPONSE_NAME);
			}

			@Test
			void shouldCallJSONService() {
				getContentValue();

				verify(jsonService).readAsListMap(any());
			}

			@Test
			void shouldReturnValueOnValidJson() {
				var expectedValue = Collections.<Map<String, Object>>emptyList();
				when(jsonService.readAsListMap(anyString())).thenReturn(expectedValue);

				var value = getContentValue();

				assertThat(value).isEqualTo(expectedValue);
			}

			@Test
			void shouldReturnEmptyListOnException() {
				when(jsonService.readAsListMap(anyString())).thenThrow(TechnicalException.class);

				var value = getContentValue();

				assertThat(value).isEmpty();
			}

			@SuppressWarnings("unchecked")
			private List<Object> getContentValue() {
				return (List<Object>) mapper.getContentValue(node, VALID_EMPTY_JSON);
			}
		}
	}

	private Map<String, Object> mapXmlToJavaMaps() {
		return mapper.mapXmlToJavaMaps(document);
	}
}
