/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.eingang.semantik.enginebased.dfoerdermittel;

import static de.ozgcloud.eingang.semantik.enginebased.dfoerdermittel.DFoerdermittelEngineBasedSemantikAdapter.*;
import static de.ozgcloud.eingang.semantik.enginebased.dfoerdermittel.DFoerdermittelFormDataTestFactory.*;
import static org.assertj.core.api.Assertions.*;
import static org.assertj.core.api.InstanceOfAssertFactories.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.util.Collections;
import java.util.Map;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;

import de.ozgcloud.common.binaryfile.TempFileUtils;
import de.ozgcloud.common.test.TestUtils;
import de.ozgcloud.eingang.common.formdata.FormData;
import de.ozgcloud.eingang.common.formdata.FormDataTestFactory;
import de.ozgcloud.eingang.common.formdata.IncomingFile;
import de.ozgcloud.eingang.common.formdata.PostfachAddressTestFactory;
import de.ozgcloud.eingang.common.formdata.ServiceKontoTestFactory;
import de.ozgcloud.eingang.common.formdata.StringBasedIdentifier;
import de.ozgcloud.eingang.common.formdata.ZustaendigeStelle;
import de.ozgcloud.eingang.common.formdata.ZustaendigeStelleTestFactory;
import de.ozgcloud.eingang.semantik.common.ServiceKontoFactory;

class DFoerdermittelEngineBasedSemantikAdapterTest {

	@Spy
	@InjectMocks
	private DFoerdermittelEngineBasedSemantikAdapter adapter;

	@Mock
	private ServiceKontoFactory serviceKontoFactory;

	@Nested
	class TestIsResponsible {

		@Test
		void shouldBeTrueWithFachnachricht() {
			var responsible = adapter.isResponsible(createWithFachnachricht());

			assertThat(responsible).isTrue();
		}

		@Test
		void shouldBeFalseForOuther() {
			var responsible = adapter.isResponsible(FormDataTestFactory.create());

			assertThat(responsible).isFalse();
		}
	}

	@Nested
	class TestParseFachnachricht {

		@Nested
		class ExtractData {
			@Test
			void shouldHaveFormData() {
				var result = adapter.extractFormDataFormXML(TestUtils.loadFile("xta/Beispieldatensatz_Fachnachricht.xml"));

				assertThat(result).isNotNull();
			}

			@Test
			void shouldHavePages() {
				var result = adapter.extractFormDataFormXML(TestUtils.loadFile("xta/Beispieldatensatz_Fachnachricht.xml"));

				assertThat(result).containsKey("Pages");
			}

			@Test
			void shouldHaveInboxRef() {
				var result = adapter.extractFormDataFormXML(TestUtils.loadFile("xta/Beispieldatensatz_Fachnachricht.xml"));

				assertThat(result).containsEntry("InboxReference", "sh/sh/4dd01647-b9d9-4775-1b50-08da3d83800a");
			}
		}

		@Nested
		class HandleFachnachrichtData {

			private final Map<String, Object> extracted = Map.of("name", "Theo");

			@Test
			void shouldCallExtractData() {
				doReturn(extracted).when(adapter).extractFormDataFormXML(any());

				adapter.parseFachnachricht(createWithFachnachricht(), createFachnachrichtFile());

				verify(adapter).extractFormDataFormXML(notNull());
			}

			@Test
			void shouldAddMap() {
				doReturn(extracted).when(adapter).extractFormDataFormXML(any());

				var result = adapter.parseFachnachricht(createWithFachnachricht(), createFachnachrichtFile());

				assertThat(result.getFormData()).containsEntry("Fachnachricht", extracted);
			}

			@Test
			void shouldIgnoreEmptyData() {
				doReturn(Collections.emptyMap()).when(adapter).extractFormDataFormXML(any());

				var result = adapter.parseFachnachricht(createWithFachnachricht(), createFachnachrichtFile());

				assertThat(result.getFormData()).doesNotContainKey("Fachnachricht");
			}
		}
	}

	FormData createWithFachnachricht() {

		return FormData.builder()
				.formData(Collections.emptyMap())
				.representation(createFachnachrichtFile()).build();
	}

	IncomingFile createFachnachrichtFile() {
		var tmpFile = TempFileUtils.writeTmpFile(TestUtils.loadFile("xta/Beispieldatensatz_Fachnachricht.xml"));

		return IncomingFile.builder()
				.file(tmpFile)
				.contentType("application/xml")
				.size(1283)
				.name("Beispieldatensatz_Fachnachricht.xml")
				.build();
	}

	@Nested
	class TestProcessFachnachricht {

		@Mock
		FormData formData0;

		@Mock
		FormData formData1;

		@Mock
		FormData formData2;

		@Mock
		FormData formData3;

		@Mock
		FormData formData4;

		@Mock
		FormData formData5;

		@BeforeEach
		void mock() {
			doReturn(formData1).when(adapter).addFormName(formData0);
			doReturn(formData2).when(adapter).addFormEngineName(formData1);
			doReturn(formData3).when(adapter).addServiceKonto(formData2);
			doReturn(formData4).when(adapter).addOrganisationsEinheitId(formData3);
			doReturn(formData5).when(adapter).addFormId(formData4);
		}

		@Test
		void shouldReturn() {
			var processedFormData = adapter.processFachnachricht(formData0);

			assertThat(processedFormData).isEqualTo(formData5);
		}
	}

	@Nested
	class TestAddServiceKonto {

		@BeforeEach
		void init() {
			when(serviceKontoFactory.buildOsiServiceKonto(any())).thenReturn(ServiceKontoTestFactory.create());
		}

		@Test
		void shouldHaveServiceKonto() {
			var formData = adapter.addServiceKonto(DFoerdermittelFormDataTestFactory.create());

			assertThat(formData.getHeader().getServiceKonto().getPostfachAddresses().getFirst().getIdentifier())
					.asInstanceOf(type(StringBasedIdentifier.class)).extracting(StringBasedIdentifier::getPostfachId)
					.isEqualTo(PostfachAddressTestFactory.POSTFACH_ID);
		}

		@Test
		void shouldRemovePrefix() {
			adapter.addServiceKonto(DFoerdermittelFormDataTestFactory.create());

			verify(serviceKontoFactory).buildOsiServiceKonto(DFoerdermittelFormDataTestFactory.POSTFACH_ID);
		}
	}

	@Nested
	class TestExtractPrefix {
		@Test
		void shouldRemoveAllBeforeLastSlash() {
			var result = adapter.extractPrefix("bla/bla/bla/12345");

			assertThat(result).isEqualTo("12345");
		}

		@Test
		void shouldBeFineWithoutSlash() {
			var result = adapter.extractPrefix("12345");

			assertThat(result).isEqualTo("12345");
		}
	}

	@Nested
	class TestAddOrganisationsEinheitId {

		@DisplayName("with no existing zustaendige stelles")
		@Nested
		class TestWithNoExistingZustaendigeStelles {
			@Test
			void shouldHaveOrganisationsEinheitId() {
				var inputFormData = DFoerdermittelFormDataTestFactory.createBuilder()
						.clearZustaendigeStelles()
						.build();

				var formData = adapter.addOrganisationsEinheitId(inputFormData);

				assertThat(formData.getZustaendigeStelles())
						.extracting(ZustaendigeStelle::getOrganisationseinheitenId)
						.containsExactly(ORGANISATIONS_EINHEIT_ID);
			}
		}

		@DisplayName("with one existing zustaendige stelle")
		@Nested
		class TestWithOneExistingZustaendigeStelle {

			@Test
			void shouldHaveOrganisationsEinheitId() {
				var formData = adapter.addOrganisationsEinheitId(DFoerdermittelFormDataTestFactory.create());

				assertThat(formData.getZustaendigeStelles())
						.extracting(ZustaendigeStelle::getOrganisationseinheitenId)
						.containsExactly(ZustaendigeStelleTestFactory.ORGANISATIONSEINHEIT_ID, ORGANISATIONS_EINHEIT_ID);
			}

			@DisplayName("should log error for existing zustaendige stelle")
			@Test
			void shouldLogErrorForExistingZustaendigeStelle() {
				adapter.addOrganisationsEinheitId(DFoerdermittelFormDataTestFactory.create());

				verify(adapter).logErrorForExistingZustaendigeStelle();
			}
		}
	}

	@Nested
	class TestAddFormName {

		@DisplayName("should have form name from title")
		@Test
		void shouldHaveFormNameFromTitle() {
			var formData = adapter.addFormName(createFormDataWithTitle(FORM_NAME));

			assertThat(formData.getHeader().getFormName()).isEqualTo(FORM_NAME);
		}

		@DisplayName("should have default form name for empty title")
		@Test
		void shouldHaveDefaultFormNameForEmptyTitle() {
			var formData = adapter.addFormName(createFormDataWithTitle(""));

			assertThat(formData.getHeader().getFormName()).isEqualTo(DEFAULT_FORM_NAME);
		}

		@DisplayName("should have default form name for missing title")
		@Test
		void shouldHaveDefaultFormNameForMissingTitle() {
			var formData = adapter.addFormName(createFormDataWithTitle(null));

			assertThat(formData.getHeader().getFormName()).isEqualTo(DEFAULT_FORM_NAME);
		}

		FormData createFormDataWithTitle(String title) {
			return createFachnachrichtBuilder().title(title).build().createFormData();
		}

	}

	@DisplayName("add form id")
	@Nested
	class TestAddFormId {

		@DisplayName("should get form id from project title")
		@Test
		void shouldGetFormIdFromProjectTitle() {
			var formData = adapter.addFormId(createFormDataWithProjectTitle(FORM_ID));

			assertThat(formData.getHeader().getFormId()).isEqualTo(FORM_ID);
		}

		@DisplayName("should keep form id if project title is empty")
		@Test
		void shouldKeepFormIdIfProjectTitleIsEmpty() {
			var formData = adapter.addFormId(createFormDataWithProjectTitle(""));

			assertThat(formData.getHeader().getFormId()).isEqualTo(DEFAULT_FORM_ID);
		}

		@DisplayName("should keep form id if project title is null")
		@Test
		void shouldKeepFormIdIfProjectTitleIsNull() {
			var formData = adapter.addFormId(createFormDataWithProjectTitle(null));

			assertThat(formData.getHeader().getFormId()).isEqualTo(DEFAULT_FORM_ID);
		}

		FormData createFormDataWithProjectTitle(String projectTitle) {
			return createFachnachrichtBuilder().projectTitle(projectTitle).build().createFormData();
		}

	}

	@Nested
	class TestAddFormEngineName {

		@Test
		void shouldHaveFormEngineName() {
			var formData = adapter.addFormEngineName(DFoerdermittelFormDataTestFactory.create());

			assertThat(formData.getHeader().getFormEngineName()).isEqualTo("dFördermittelantrag");
		}

	}
}
