/*
 * Copyright (C) 2023 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.eingang.semantik.enginebased.formsolutions;

import static de.ozgcloud.eingang.semantik.enginebased.formsolutions.FormSolutionsHeaderTestFactory.*;
import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;

import de.ozgcloud.eingang.common.formdata.FormData;
import de.ozgcloud.eingang.common.formdata.FormDataUtils;
import de.ozgcloud.eingang.common.formdata.FormHeader;
import de.ozgcloud.eingang.semantik.common.ServiceKontoFactory;

class FormSolutionsHeaderMapperTest {

	@Spy
	@InjectMocks
	private final FormSolutionsHeaderMapper mapper = new FormSolutionsHeaderMapper();
	@Mock
	private ServiceKontoFactory serviceKontoFactory;

	@DisplayName("Parse formData")
	@Nested
	class TestParseFormData {

		private final FormData formData = FormSolutionsHeaderTestFactory.create();

		@Test
		void shouldCallBuildFormHeader() {
			parseFormData();

			verify(mapper).buildFormHeader(formData);
		}

		@Test
		void shouldReturnValue() {
			var result = parseFormData();

			assertThat(result).usingRecursiveComparison().ignoringFields("header", "formData").isEqualTo(formData);
		}

		@Test
		void shouldRemoveTransactionId() {
			var result = parseFormData();

			assertThat(result.getFormData()).doesNotContainKey(FormSolutionsHeaderMapper.TRANSACTION_ID);
		}

		private FormData parseFormData() {
			return mapper.parseFormData(formData);
		}

		@DisplayName("build form header")
		@Nested
		class TestBuildFormHeader {

			@Test
			void shouldHaveFormName() {
				var formHeader = buildFormHeader();

				assertThat(formHeader.getFormName()).isEqualTo(FORM_NAME);
			}

			@Test
			void shouldHaveFormId() {
				var formHeader = buildFormHeader();

				assertThat(formHeader.getFormId()).isEqualTo(FORM_NAME);
			}

			@Test
			void shouldHaveTransactionId() {
				var formHeader = buildFormHeader();

				assertThat(formHeader.getRequestId()).isEqualTo(REQUEST_ID);
			}

			@Test
			void shouldHaveFormEngineName() {
				var formHeader = buildFormHeader();

				assertThat(formHeader.getFormEngineName()).isEqualTo(FormSolutionsHeaderMapper.FORM_ENGINE_NAME);
			}

			@DisplayName("service konto")
			@Nested
			class TestGetServiceKonto {

				@Test
				void shouldCallServiceKontoBuildHelper() {
					buildFormHeader();

					verify(serviceKontoFactory).buildOsiServiceKonto(any());
				}

				@Test
				void shouldNotCallServiceKontoBuildHelper() {
					var formDataWithoutPostkorbHandle = FormDataUtils.from(formData).remove(FormSolutionsHeaderMapper.POSTKORBHANDLE).build();

					mapper.buildFormHeader(formDataWithoutPostkorbHandle);

					verify(serviceKontoFactory, never()).buildOsiServiceKonto(any());
				}

			}

			private FormHeader buildFormHeader() {
				return mapper.buildFormHeader(formData);
			}
		}
	}
}