/*
 * Copyright (C) 2023 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.eingang.semantik.enginebased.formsolutions;

import static de.ozgcloud.eingang.semantik.enginebased.formsolutions.FormSolutionsEngineBasedAdapter.*;
import static de.ozgcloud.eingang.semantik.enginebased.formsolutions.FormSolutionsPanelTestFactory.*;
import static org.assertj.core.api.Assertions.*;

import java.util.Map;

import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;

import de.ozgcloud.eingang.common.formdata.FormData;
import de.ozgcloud.eingang.semantik.enginebased.formsolutions.FormSolutionsPanelMapper;

class FormSolutionsPanelMapperTest {

	private FormSolutionsPanelMapper mapper = new FormSolutionsPanelMapper();

	@Nested
	class TestMapping {

		@Test
		void shouldMap() {
			var formData = mapper.parseFormData(FormSolutionsPanelTestFactory.create());

			assertThat(formData.getFormData()).isNotEmpty();
		}
	}

	@Nested
	class TestMapPanel {
		@Test
		void shouldContainIdentifier() {
			var formData = mapper.parseFormData(FormSolutionsPanelTestFactory.create());

			assertThat(formData.getFormData()).containsKey(PANEL_0);
		}

		@Test
		@SuppressWarnings("unchecked")
		void shouldContainComponents() {
			var formData = mapper.parseFormData(FormSolutionsPanelTestFactory.create());

			var panelContent = (Map<String, Object>) formData.getFormData().get(PANEL_0);

			assertThat(panelContent).containsEntry(DATE_FIELD, DATE_VALUE_CONTENT).containsEntry(TEXT_FIELD, STRING_VALUE_CONTENT);
		}

		@Test
		void shouldContainNestedComponents() {
			var formData = mapper.parseFormData(FormSolutionsPanelTestFactory.createBuilder().formData(NESTED_PANEL_FORM).build());

			assertThat(formData.getFormData()).containsKey(PANEL_0);
		}
	}

	@Nested
	class TestMapComponents {
		@Nested
		class TestMapNestedComponents {

			@Test
			@SuppressWarnings("unchecked")
			void shouldHaveGroup() {
				var formData = mapper.parseFormData(FormSolutionsPanelTestFactory.createBuilder().formData(NESTED_PANEL_FORM).build());

				var group = (Map<String, Object>) formData.getFormData().get(PANEL_0);

				assertThat(group).containsKey(GROUP_IDENTIFIER);
			}

			@Test
			@SuppressWarnings("unchecked")
			void shouldHaveComponent() {
				var formData = mapper.parseFormData(FormSolutionsPanelTestFactory.createBuilder().formData(NESTED_PANEL_FORM).build());

				var group = (Map<String, Object>) ((Map<String, Object>) formData.getFormData().get(PANEL_0)).get(GROUP_IDENTIFIER);

				assertThat(group).containsEntry(DATE_FIELD, DATE_VALUE_CONTENT);
			}

			@Test
			@SuppressWarnings("unchecked")
			void shouldNotHaveEmptyComponent() {
				var formData = mapper.parseFormData(FormSolutionsPanelTestFactory.createBuilder().formData(PANEL_FORM_EMPTY).build());

				var group = (Map<String, Object>) ((Map<String, Object>) formData.getFormData().get(PANEL_0)).get(TEXT_FIELD);

				assertThat(group).isNull();
			}
		}
	}

	@Nested
	class TestGetPanels {

		@Test
		void shoudReturnPanels() {
			var panels = FormSolutionsPanelMapper.getPanels(FormSolutionsPanelTestFactory.create());

			assertThat(panels).isEqualTo(PANEL_LIST);
		}

		@Test
		@DisplayName("should return empty list when ASSISTANT map is missing")
		void shouldHandleMissingAssistant() {
			var formData = FormSolutionsPanelTestFactory.createBuilder().formData(Map.of()).build();

			var panels = FormSolutionsPanelMapper.getPanels(formData);

			assertThat(panels).isEmpty();
		}

		@Test
		@DisplayName("should return empty list when PANELS map is missing")
		void shouldHandleMissingPanels() {
			var formData = FormSolutionsPanelTestFactory.createBuilder().formData(Map.of(ASSISTANT, Map.of())).build();

			var panels = FormSolutionsPanelMapper.getPanels(formData);

			assertThat(panels).isEmpty();
		}

		@Test
		void shouldHandleNullFormDataMap() {
			var panels = FormSolutionsPanelMapper.getPanels(FormData.builder().build());

			assertThat(panels).isEmpty();
		}

		@Test
		void shouldHandleNullFormData() {
			var panels = FormSolutionsPanelMapper.getPanels(null);

			assertThat(panels).isEmpty();
		}
	}

	@Nested
	class TestGetComponents{

		@Test
		void shouldReturnComponentList() {
			var components = FormSolutionsPanelMapper.getComponentList(PANEL_LIST.get(0));

			assertThat(components).isEqualTo(COMPONENT_LIST);
		}

		@Test
		void shouldHandleNull() {
			var components = FormSolutionsPanelMapper.getComponentList(null);

			assertThat(components).isEmpty();
		}

		@Test
		void shouldHandleEmptyMap() {
			var components = FormSolutionsPanelMapper.getComponentList(Map.of());

			assertThat(components).isEmpty();
		}
	}
}
